/*  Copyright (c) 1995 John E. Davis (davis@space.mit.edu)
 *  All rights reserved.
 */
#include <config.h>

#include <stdio.h>
#include <signal.h>
#include <string.h>


#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef VMS
# include <unixlib.h>
# include <unixio.h>
# include <dvidef.h>
# include <descrip.h>
#else
# ifndef sun
#  include <sys/ioctl.h>
# endif
# ifdef HAVE_TERMIOS_H
#  include <termios.h>
# endif
# ifdef SYSV
#  include <sys/termio.h>
#  include <sys/stream.h>
#  include <sys/ptem.h>
#  include <sys/tty.h>
# endif
#endif /* !VMS */

#include <slang.h>



#include "clientlib.h" 
#include "nntp.h"
#include "slrn.h"
#include "group.h"
#include "misc.h"
#include "startup.h"
#include "art.h"
#include "score.h"

#ifdef __os2__
# define INCL_VIO
# include <os2.h>
#endif

#ifndef NNTPSERVER_FILE
# define NNTPSERVER_FILE NULL
#endif

#ifdef SIGSTOP
# define USE_RESUME_SMG
#endif


int Slrn_TT_Initialized = 0;

/* If -1, force mouse.  If 1 the mouse will be used on in XTerm.  If 0, 
 * do not use it.
 */
int Slrn_Use_Mouse;

int Slrn_Suspension_Ok;
void slrn_set_suspension (int ok)
{
   if (ok == Slrn_Suspension_Ok) return;
#ifdef USE_RESUME_SMG
   SLtty_set_suspend_state (ok);
#endif
   Slrn_Suspension_Ok = ok;
}

void slrn_smg_refresh (void)
{
   int ok = Slrn_Suspension_Ok;
   if (ok) slrn_set_suspension (0);
   SLsmg_refresh ();
   if (ok) slrn_set_suspension (1);
}

void slrn_suspend (int sig)
{
#ifdef SIGSTOP
   int init = Slrn_TT_Initialized;
   slrn_reset_display (sig);
   kill(getpid(),SIGSTOP);
   slrn_init_display (init, sig);
# ifdef USE_RESUME_SMG
   if (Slrn_Suspension_Ok) SLtty_set_suspend_state (1);
# endif
#else
   slrn_error ("Not implemented.");
#endif
}

static void lock_file (int how)
{
   char file[1024];
   char name[256];
   char buf[80];
   static int not_ok_to_unlock;
   FILE *fp;
   
   if (Slrn_Newsrc_File == NULL) return;
   if (not_ok_to_unlock) return;
   
   not_ok_to_unlock = 1;
   sprintf (name, "%s-lock", Slrn_Newsrc_File);
   slrn_make_home_filename (name, file);
   
   if (how == 1)
     {
	int pid;
	if ((fp = fopen (file, "r")) != NULL)
	  {
	     if (NULL != fgets (buf, 79, fp))
	       {
		  sscanf (buf, "%d", &pid);
		  if ((pid > 0) && (0 == kill (pid, 0)))
		    {
		       slrn_exit_error ("slrn: pid %d is locking the newsrc file.", pid);
		    }
	       }
	     slrn_fclose (fp);
	  }
#ifdef VMS
	fp = fopen (file, "w", "fop=cif");
#else
	fp = fopen (file, "w");
#endif
	if (NULL == fp)
	  {
	     slrn_exit_error ("Unable to create lock file %s.", file);
	  }
	pid = getpid ();
        fprintf (fp, "%d", pid);
	slrn_fclose (fp);
     }
   else
     {
	if (-1 == slrn_delete_file (file))
	  {
	     /* slrn_exit_error ("Unable to remove lockfile %s.", file); */
	  }
     }
   not_ok_to_unlock = 0;
}

void (*Slrn_Winch_Function)(void);

static void slrn_set_screen_size (int sig)
{
   int r = 0, c = 0;
#ifdef VMS
   int status, code;
   unsigned short chan;
   $DESCRIPTOR(dev_dsc, "SYS$INPUT:");
#endif
#ifdef __os2__
   VIOMODEINFO vioModeInfo;
#endif
   
#ifdef TIOCGWINSZ
   struct winsize wind_struct;
   
   if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
       || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
       || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
     {
	c = (int) wind_struct.ws_col;
	r = (int) wind_struct.ws_row;
     }
   
#endif
#ifdef VMS
   status = sys$assign(&dev_dsc,&chan,0,0,0);
   if (status & 1)
     {
	code = DVI$_DEVBUFSIZ;
	status = lib$getdvi(&code, &chan,0, &c, 0,0);
	if (!(status & 1))
	  c = 80;
	code = DVI$_TT_PAGE;
	status = lib$getdvi(&code, &chan,0, &r, 0,0);
	if (!(status & 1))
	  r = 24;
	sys$dassgn(chan);
     }
#endif

#ifdef __os2__
   vioModeInfo.cb = sizeof(vioModeInfo);
   VioGetMode (&vioModeInfo, 0);
   c = vioModeInfo.col;
   r = vioModeInfo.row;   
#endif
   
   if (r <= 0)
     {
	char *s = getenv ("LINES");
	if (s != NULL) r = atoi (s);
     }
   
   if (c <= 0)
     {
	char *s = getenv ("COLUMNS");
	if (s != NULL) c = atoi (s);
     }
   
   if ((r <= 0) || (r > 200)) r = 24;
   if ((c <= 0) || (c > 250)) c = 80;
   SLtt_Screen_Rows = r;
   SLtt_Screen_Cols = c;
   
   Group_Window_Size = SLtt_Screen_Rows - 3;
   
#ifdef SIGWINCH
   if (sig == SIGWINCH)
     {
	SLsmg_reset_smg ();
	SLsmg_init_smg ();
	if (Slrn_Winch_Function != NULL) (*Slrn_Winch_Function) ();
	slrn_redraw ();
     }
   signal (SIGWINCH, slrn_set_screen_size);
#endif
}


void (*Slrn_Hangup_Hook) (void);
static void slrn_hangup (int sig)
{
   if (Slrn_Hangup_Hook != NULL) (*Slrn_Hangup_Hook) ();
   exit (-1);
}

void slrn_enable_mouse (int mode)
{
#ifndef __os2__
   if (Slrn_Use_Mouse)
     {
	if (-1 == SLtt_set_mouse_mode (mode, (Slrn_Use_Mouse < 0)))
	  Slrn_Use_Mouse = 0;
     }
#endif
}

/* if 1, init just slsmg.  If 2, just sltt, if 3, do both */
void slrn_init_display (int init_mode, int sig)
{
   Slrn_TT_Initialized |= init_mode;
   
#ifdef SIGTSTP
   signal(SIGTSTP, slrn_suspend);
#endif
   
#ifdef SIGPIPE
   signal(SIGPIPE, SIG_IGN);
#endif
#ifdef SIGTTOU
   /* Allow background writes */
   signal (SIGTTOU, SIG_IGN);
#endif
   signal (SIGHUP, slrn_hangup);
#ifdef SIGTERM
   signal (SIGTERM, slrn_hangup);
#endif
   
   if (init_mode & 2)
     {
	SLang_init_tty (7, 1, 0);
	SLang_set_abort_signal (NULL);
     }
   
   if (init_mode & 1)
     {
	slrn_enable_mouse (1);
#ifdef USE_RESUME_SMG
	if (sig == SIGTSTP) SLsmg_resume_smg ();
	else
	  {
#endif
	     slrn_set_screen_size (0);
	     SLsmg_init_smg ();
	     
	     /* We do not want the -> overlay cursor to affect the scroll. */
#ifndef __os2__
	     SLsmg_Scroll_Hash_Border = 2;
#endif
	     slrn_redraw ();
	  }
#ifdef USE_RESUME_SMG
     }
#endif
}

void slrn_reset_display (int sig)
{
   int init = Slrn_TT_Initialized;
   
   if (init == 0) return;
   
   slrn_enable_mouse (0);
   
#ifdef USE_RESUME_SMG
   if (sig == SIGTSTP) SLsmg_suspend_smg ();
   else
     {
#endif
	SLsmg_gotorc (SLtt_Screen_Rows - 1, 0);
	slrn_smg_refresh ();
	SLsmg_reset_smg ();
#ifdef SIGWINCH
	signal (SIGWINCH, SIG_DFL);
#endif
#ifdef USE_RESUME_SMG
     }
#endif
   
   if (init & 2) SLang_reset_tty ();
   
#ifdef SIGTSTP
   signal(SIGTSTP, SIG_DFL);
#endif
   
   Slrn_TT_Initialized = 0;
}

char *Slrn_Server_Name;

void slrn_tty_message (int newline_flag, char *fmt, ...)
{
   va_list ap;
   
   if (newline_flag & 1) putc ('\n', stdout);
   va_start (ap, fmt);
   vfprintf (stdout, fmt, ap);
   va_end(ap);
   if (newline_flag & 2) putc ('\n', stdout);
   fflush (stdout);
}


int slrn_get_new_news (int no_new_groups, int create_flag)
{
   char msg1[] = "Checking for new groups...";
   char msg2[] = "Checking news...";
   char msg3[] = "Checking news via active file...";
   char *msg;
   
   if (nntp_initialize_server (Slrn_Server_Name) != 0) return (-1);
   
   if (create_flag == 0)
     {
	if (no_new_groups == 0)
	  {
	     if (Slrn_TT_Initialized & 1)
	       {
		  slrn_message (msg1);
		  slrn_smg_refresh ();
	       }
	     else
	       slrn_tty_message (1, msg1);
	     
	     slrn_check_new_groups (create_flag);
	  }
	
	if (Slrn_List_Active_File) msg = msg3;
	else msg = msg2;
	
	if (Slrn_TT_Initialized & 1)
	  {
	     slrn_message (msg);
	     slrn_smg_refresh ();
	  }
	else
	  slrn_tty_message (1, msg);
     }
   slrn_read_newsrc (create_flag);
   
   slrn_read_group_descriptions ();
   
   return (0);
}

char *Slrn_Newsrc_File = NULL;
char *Slrn_Lib_Dir = NULL;

static void usage (char *extra)
{
   fprintf (stderr, "slrn (%s):\n\
Usage: slrn [-n] [-C] [-h nntp-host]  [-f newsrc-file] [-create]\n\
-n              Do not check for new groups.  This usually results in\n\
                 a faster startup.\n\
-h nntp-host    Host name to connect to.  This overrides NNTPSERVER variable.\n\
-f newsrc-file  Name of the newsrc file to use.\n\
-C              Use colors.\n\
-create         Create a newsrc file by getting list of groups from server.\n\
-d              Get new text descriptions of each group from server.\n\
                 Note: This may take a LONG time to retrieve this information.\n\
                 The resulting file can be several hundred Kilobytes!\n\
-i init-file    Name of initialization file to use (default .slrnrc)\n\
-Dname          Add 'name' to list of predefined preprocessing tokens.\n\
-p NNTP-PORT    Set the NNTP port to NNTP-PORT. The default value is 119.\n\
                 Note: This option has no effect on some systems.\n\
-k              Do not process score file.\n\
-k0             Process score file but inhibit expensive scores.\n\
-a              Use active file for getting new news\n\
-m              Force XTerm mouse reporting\n\
-help           Print this usage.\n",
	    Slrn_Version);
   
   if (extra != NULL)
     {
	fprintf (stderr, "\n%s\n", extra);
     }
   
   exit (-1);
}

static char *port_change (char *hostname)
{
   char *colon;
   char *h;
   
   colon = slrn_strchr (hostname, ':');
   if (colon == NULL) return hostname;
   
   h = SLmake_string (hostname);
   if (h == NULL)
     slrn_exit_error ("Out of memory.");
   
   colon = h + (colon - hostname);
   *colon++ = 0;
   Slrn_NNTP_Port = atoi (colon);
   return h;
}

int main (int argc, char **argv)
{
   char *host = NULL, *hlp_file;
   int i;
   int create_flag = 0;
   int no_new_groups = 0;
   int no_score_file = 0;
   int use_color = 0;
   int use_mouse = 0;
   int dsc_flag = 0;
   FILE *fp;
   char file[512];
#ifdef __os2__
   char *init_file = "slrn.rc";
#else
   char *init_file = ".slrnrc";
#endif
   
   
#ifdef unix
   SLdefine_for_ifdef ("UNIX");
#endif
#ifdef VMS
   SLdefine_for_ifdef ("VMS");
#endif
   
   for (i = 1; i < argc; i++)
     {
	if (!strcmp ("-create", argv[i])) create_flag = 1;
	else if (!strcmp ("-C", argv[i])) use_color = 1;
	else if (!strcmp ("-a", argv[i])) Slrn_List_Active_File = 1;
	else if (!strcmp ("-n", argv[i])) no_new_groups = 1;
	else if (!strcmp ("-d", argv[i])) dsc_flag = 1;
	else if (!strcmp ("-m", argv[i])) use_mouse = 1;
	else if (!strcmp ("-k", argv[i])) no_score_file = 1;
	else if (!strcmp ("-k0", argv[i]))
	  Slrn_Perform_Scoring &= ~SLRN_EXPENSIVE_SCORING;
	
	else if (!strncmp ("-D", argv[i], 2) && (argv[i][2] != 0))
	  {
	     if (SLdefine_for_ifdef (argv[i] + 2) == 0)
	       {
		  slrn_exit_error ("Unable to add preprocessor name %s.",
				   argv[i] + 2);
	       }
	  }
	else if (i + 1 < argc)
	  {
	     if (!strcmp("-h", argv[i])) host = argv[++i];
	     else if (!strcmp ("-f", argv[i])) Slrn_Newsrc_File = argv[++i];
	     else if (!strcmp ("-i", argv[i])) init_file = argv[++i];
	     else if (!strcmp ("-p", argv[i]))
	       {
		  Slrn_NNTP_Port = atoi (argv[++i]);
	       }
	     else usage (NULL);
	  }
	else usage (NULL);
     }

   fprintf (stdout, "slrn %s\n", Slrn_Version);

   if (dsc_flag && create_flag)
     {
	usage ("The -d and -create flags must not be specified together.");
     }
   
   if ((host == NULL) && (NULL == (host = getserverbyfile(NNTPSERVER_FILE))))
     {
	fprintf (stderr, "You need to set the NNTPSERVER environment variable to your server name.\r\n");
#ifdef VMS
	fprintf (stderr, "Example: $ define/job NNTPSERVER my.news.server\r\n");
#else
# ifdef __os2__
	fprintf (stderr, "Example: set NNTPSERVER=my.news.server\r\n");
# else
	fprintf (stderr, "Example (csh): setenv NNTPSERVER my.news.server\r\n");
# endif
#endif
	exit (-1);
     }
   
   Slrn_Server_Name = port_change (host);
   
   SLtt_get_terminfo();
   if (use_color) SLtt_Use_Ansi_Colors = 1;
   slrn_initialize ();
   slrn_get_user_info ();
   
   if (Slrn_Lib_Dir == NULL)
     {
	Slrn_Lib_Dir = getenv ("SLRN_LIB_DIR");
     }
   if (Slrn_Lib_Dir == NULL) Slrn_Lib_Dir = SLRN_LIB_DIR;
   
#ifdef VMS
   sprintf (file, "%s%s", Slrn_Lib_Dir, "slrn.rc");
#else
   sprintf (file, "%s/%s", Slrn_Lib_Dir, "slrn.rc");
#endif
   
   slrn_read_startup_file (file);      /* global file for all users */
   slrn_read_startup_file (init_file);
   
   /* Now that we have read in the startup file, check to see if the user
    * has a username and a usable hostname.  Without those, we are not
    * starting up.
    */
   if (0 == slrn_is_fqdn (Slrn_User_Info.host))
     {
	slrn_exit_error ("\
Unable to find a fully qualified hostname.  You will have to specify a\r\n\
hostname in your .slrnrc file (slrn.rc on VMS).\r\n");
     }
   if ((NULL == Slrn_User_Info.username)
       || (0 == *Slrn_User_Info.username)
       || (NULL != slrn_strchr (Slrn_User_Info.username, '@')))
     {
	slrn_exit_error ("\
Unable to find you user name.  This means that a valid 'From' header line\r\n\
cannot be constructed.  Try setting the USER environment variable.\r\n");
     }
   
   if ((no_score_file == 0) && (Slrn_Score_File != NULL)
       && (-1 == slrn_read_score_file (Slrn_Score_File)))
     {
	slrn_exit_error ("Error processing score file %s", Slrn_Score_File);
     }
   
   hlp_file = getenv ("SLRNHELP");
   if (hlp_file != NULL)
     {
	slrn_parse_helpfile (hlp_file);
     }
   
   if ((Slrn_Newsrc_File == NULL)
       && ((Slrn_Newsrc_File = slrn_map_file_to_host (host)) == NULL))
#if defined(VMS) || defined(__os2__)
     Slrn_Newsrc_File = "jnews.rc";
#else
   Slrn_Newsrc_File = ".jnewsrc";
#endif

   if (use_mouse) Slrn_Use_Mouse = -1; /* -1 forces it. */
   
   if (dsc_flag)
     {
	if (nntp_initialize_server (Slrn_Server_Name) != 0)
	  {
	     slrn_exit_error ("Unable to initialize server.");
	     return (-1);
	  }
	slrn_get_group_descriptions ();
	nntp_close_server ();
	return 0;
     }
   
   if (create_flag == 0)
     {
	/* Check to see if the .newrc file exists--- I should use the access
	 * system call but for now, do it this way.
	 */
	if (NULL == (fp = slrn_open_home_file (Slrn_Newsrc_File, "r", file, 0)))
	  {
	     slrn_error ("Unable to open %s.  I will try .newsrc.", file);
	     if (NULL == (fp = slrn_open_home_file (".newsrc", "r", file, 0)))
	       {
		  slrn_exit_error ("\r\nUnable to open %s.\r\n\
If you want to create %s, use the command:\r\n\
slrn -f %s -create\r\n", file, file, file);
	       }
	  }
	slrn_fclose (fp);
	
	lock_file (1);
     }
   
   if (-1 == slrn_get_new_news (no_new_groups, create_flag))
     {
	slrn_exit_error ("Failed to initialize server.");
     }
   
   putc ('\n', stdout);
   slrn_init_display (1 | 2, 0);
#if defined(unix) && !defined(__os2__)
   if (Slrn_Autobaud) SLtt_Baud_Rate = SLang_TT_Baud_Rate;
#endif
   
   slrn_group_mode ();
   
   slrn_quit (0);
   return 0;
}

void slrn_quit (int retcode)
{
   nntp_close_server ();
   slrn_reset_display (0);
   lock_file (0);
   exit (retcode);
}


void slrn_exit_error (char *fmt, ...)
{
   va_list ap;
   
   if (Slrn_TT_Initialized)
     {
	if (Slrn_Groups_Dirty) slrn_write_newsrc ();
	slrn_reset_display (0);
	client_close_server ();
     }
   
   va_start (ap, fmt);
   vfprintf (stderr, fmt, ap);
   va_end(ap);
   putc ('\n', stderr);
   lock_file (0);
   exit (-1);
}
